﻿#Region "FileHeader"
#If False Then
FigureDataTable ist fast trivial. Des Pudels Kern ist, dass bei FigureRowChanged e.Row.ApplyChanges() aufgerufen wird, um die anhängige Grafik der FigureRow zu aktualisieren.
FigureRow.ApplyChanges ist komplexer, da verschiedene Änderungen aufgetreten sein können: die Location kann geändert sein, oder auch die Location eines Grips, oder die Color.
Ausserdem ist zu berücksichtigen, ob die Row selected ist.
ApplyChanges geht prinzipiell 3-schrittig vor: 
1) die Bounds werden invalidiert. Das verursacht kurz darauf die Löschung der bisherigen Zeichnung. 
2) Die Änderungen werden verarbeitet. Konkret: die Farbe wird gesetzt, und der GraphicsPath wird mit den absoluten Koordinaten aller Grips gefüttert. Als letztes die Bounds dieses neuen GraphicPathes merken.
3) die (neuen) Bounds auch invalidieren. Das verursacht darauf die Neu-Zeichnung
Die Zeichnung selbst wird nicht in ApplyChanges ausgeführt, sondern dafür ist Draw() zuständig, wo der in ApplyChanges vorbereitete GP nun jederzeit und immer wieder gedrawt werden kann.

Also die Invalidate-Logik ist: der invalidierte Bereich wird kurz darauf neu gezeichnet. Mehrere Invalidates fliessen in einen beide umfassenden Invalid-Bereich zusammen. Sodass der kurz darauf folgende Zeichen-Vorgang beides erledigt: Löschen sowieso, und zeichnen kann er ja garnet anders als an der neuen Position.

Komplex auch die AddGrip-Methode: alle Grips werden ermittelt, nach Index sortiert. Nach dem aktuell gehoverten Grip werden alle Indizees erhöht, und dann kann ein neuer Grip geadded werden, mit einem Index, der die Lücke auffüllt. Lokalisiert wird der Neue genau zwischen seine Nachbarn, und ausserdem gleich selected.
Während dieser "Massen-Operation" wird die Grip-Logik deaktiviert, und zwar nicht nur die innere Logik der Grip-Klasse, sondern auch alle BindingSources, die GripRows anzeigen, werden ruhiggestellt (DataTableAttach.SuspendLevel.BindingSources)
#End If '-- Options, Imports
Imports System.Drawing.Drawing2D
#End Region 'FileHeader

Partial Class FigureDts

   Partial Class FigureDataTable

      Private WithEvents _Attach As DataTableAttach = Me.Attach
      Private WithEvents _Active As FigureDataTable = Nothing
      Private Sub _Attach_Suspending(ByVal sender As Object, ByVal e As DataTableAttach.SuspendingEventArgs) Handles _Attach.Suspending
         e.TriDecide(_Active, Nothing, Me)
      End Sub

      Private Sub _Active_FigureRowChanged(sender As Object, e As FigureRowChangeEvent) Handles _Active.FigureRowChanged
         Select Case e.Action
            Case DataRowAction.Add ' 3 Standard-Grips rein-generieren
               Using Dts.Grip.Attach.BeginSuspend(DataTableAttach.Suspend.NotifyOnly)
                  With Dts.Grip
                     .AddGripRow(0, 0, e.Row, 0)
                     .AddGripRow(50, 0, e.Row, 1)
                     .AddGripRow(25, 36, e.Row, 2)
                  End With
               End Using
            Case DataRowAction.Change
            Case Else : Return
         End Select
         e.Row.ApplyChanges() ' Grafik updaten
      End Sub

      Private Sub _Active_FigureRowDeleted(sender As Object, e As FigureRowChangeEvent) Handles _Active.FigureRowDeleted
         Dts.Invalidate(e.Row.Bounds, _HoverPen)
      End Sub

      Private Sub _Active_FigureRowDeleting(sender As Object, e As FigureRowChangeEvent) Handles _Active.FigureRowDeleting
         'falls ein Grip/Figure - Hover- oder Select-Markup von der Löschung betroffen ist, es Null-setzen
         For Each grp In e.Row.GetGripRows
            Dts.GripHover.ClearMarkup(grp)
            Dts._GripSelect.ClearMarkup(grp)
         Next
         Dts._FigureSelect.ClearMarkup(e.Row)
         Dts._FigureHover.ClearMarkup(e.Row)
      End Sub

      Private Sub _Active_TableNewRow(sender As Object, e As DataTableNewRowEventArgs) Handles _Active.TableNewRow
         With DirectCast(e.Row, FigureRow)
            .Name = "Figure" & -.ID ' Standard-Name vorbelegen
         End With
      End Sub

   End Class

   Partial Class FigureRow

      Public ReadOnly Pen As New Pen(Brushes.AliceBlue) With {.LineJoin = LineJoin.Round}
      Public ReadOnly SelectedPen As New Pen(Brushes.AliceBlue, 3) With {.LineJoin = LineJoin.Round}
      Public ReadOnly Geometry As New GraphicsPath
      Public Bounds As Rectangle

      Public Sub ApplyChanges()
         Dim pn = If(Dts._FigureSelect.IsMarked(Me), SelectedPen, Pen)
         Dts.Invalidate(Bounds, pn)
         Geometry.Reset()
         Pen.Color = Color
         SelectedPen.Color = Color
         Dim pt As Point = Location
         Geometry.AddLines(Aggregate rw In GetGripRows() Order By rw.Index Into ToArray(pt + rw.Location))
         Geometry.CloseFigure()
         Bounds = Geometry.GetBounds.RoundX(True)
         Dts.Invalidate(Bounds, pn)
      End Sub

      Public Property Color As Color
         Get
            Return Color.FromArgb(ColorCode)
         End Get
         Set(value As Color)
            Dim cc = value.ToArgb
            If ColorCode = cc Then Return
            ColorCode = cc
         End Set
      End Property

      ''' <summary> das die Zeichnung umschließende Rechteck, bei Stiftbreite 0. Beachte: die tatsächliche Zeichnung ist um die Stiftbreite größer! </summary>
      Public Property Location As Vector
         Get
            Return New Vector(X, Y)
         End Get
         Set(value As Vector)
            If value = Location Then Return
            Using Dts.Figure.Attach.BeginSuspend(DataTableAttach.Suspend.NotifyOnly) ' Figure_RowChanged-Event suspendieren
               With value
                  X = .X : Y = .Y ' ansonsten würden 2 Zuweisungen 2 _FigureRowChanged() auslösen
               End With
            End Using
            ApplyChanges() ' nach Daten-Änderung den Grafik-Zustand aktualisieren, und Anzeige auslösen
         End Set
      End Property

      Public Sub Draw(g As Graphics, Optional pen As Pen = Nothing)
         g.DrawPath(If(pen, Me.Pen), Geometry)
      End Sub

      Public Function Contains(pt As Point) As Boolean
         Return Geometry.IsOutlineVisible(pt, _HoverPen)
      End Function

      Public Sub AddGrip()
         Dim tbGrip = Dts.Grip
         Dim grips = GetGripRows().OrderBy(Function(rw) rw.Index).ToArray
         Dim indx = Dts.GripHover.Item.Index
         Dim nxt = grips((indx + 1) Mod grips.Length)
         Dim newLocation = (Dts.GripHover.Item.Location + nxt.Location) \ 2
         Using tbGrip.Attach.BeginSuspend(DataTableAttach.Suspend.NotifyOnly)
            For i = indx + 1 To grips.Length - 1
               grips(i).Index += 1
            Next
         End Using
         Dts._FigureSelect.Item = Me
         Dts._GripSelect.Item = tbGrip.AddGripRow(newLocation.X, newLocation.Y, Me, indx + 1)
      End Sub

   End Class

End Class
