﻿Public Class GraphicMarkup(Of T As Class)

   Public Draw As Action(Of T, Graphics, Pen)
   Public Invalidate As Action(Of Rectangle, Pen)
   Public GetBounds As Func(Of T, Rectangle)
   Public GetPen As Func(Of T, Pen)

   Public Event ItemChanged As EventHandler
   Public Sub New()
   End Sub
   Public Sub New(gm As GraphicMarkup(Of T))
      With gm
         Draw = .Draw
         Invalidate = .Invalidate
         GetBounds = .GetBounds
         GetPen = .GetPen
      End With
   End Sub

   Private _Item As T
   Public Property Item() As T
      Get
         Return _Item
      End Get
      Set(value As T)
         If value Is _Item Then Return
         TryInvalidate()
         _Item = value
         TryInvalidate()
         RaiseEvent ItemChanged(Me, EventArgs.Empty)
      End Set
   End Property

   Public ReadOnly Property HasMarkup As Boolean
      Get
         Return _Item.NotNull
      End Get
   End Property
   ''' <summary> clears markup. optional only if the passed itm is the marked one </summary>
   Public Sub ClearMarkup(Optional itm As T = Nothing)
      If itm Is Nothing OrElse itm Is _Item Then Item = Nothing
   End Sub
   Public Function IsMarked(itm As Object) As Boolean
      Return itm Is _Item
   End Function

   Private Sub TryInvalidate()
      If _Item.Null Then Return
      Invalidate(GetBounds(_Item), GetPen(_Item))
   End Sub

   Public Function TryDraw(g As Graphics) As Boolean
      TryDraw = _Item.NotNull
      If TryDraw Then Draw(_Item, g, GetPen(_Item))
   End Function

End Class
