﻿Imports System.Text
Imports System.Globalization
Imports System.ComponentModel

Public Class ControlBehavior

   Private Shared _DroppingCombo As DataGridViewComboBoxEditingControl
   Private Shared Sub Application_Idle(ByVal sender As Object, ByVal e As EventArgs)
      RemoveHandler Application.Idle, AddressOf Application_Idle
      _DroppingCombo.DroppedDown = True
   End Sub

   Public Shared Sub Grid_DataError(ByVal sender As Object, ByVal e As DataGridViewDataErrorEventArgs)
      Dim grd = DirectCast(sender, DataGridView)
      Dim msg = String.Join(Lf, _
         "Eine gravierende Fehl-Eingabe wurde festgestellt (wie etwa eine Text-Eingabe in ein Zahlenfeld)", _
         "", _
         "Fehler: " & e.Exception.Message)
      MessageBox.Show(grd.FindForm, msg, "Eingabe abgelehnt")
      grd.CancelEdit()
   End Sub

   Public Shared Sub Grid_FocusChanged(ByVal sender As Object, ByVal e As EventArgs)
      With DirectCast(sender, DataGridView)
         .DefaultCellStyle.SelectionBackColor = If(.Focused, Color.Khaki, Color.FromArgb(240, 240, 180))
      End With
   End Sub

   Public Shared Sub CurrencyGrid_CellValidating(ByVal sender As Object, ByVal e As DataGridViewCellValidatingEventArgs)
      Dim cell = DirectCast(sender, DataGridView)(e.ColumnIndex, e.RowIndex)
      If cell.Value Is Nothing OrElse Not cell.ValueType.Equals(GetType(Decimal)) Then Return
      Dim num As Decimal
      e.Cancel = Not Decimal.TryParse(e.FormattedValue.ToString, NumberStyles.Currency, CultureInfo.CurrentCulture, num) OrElse num Mod 0.01D <> 0
      If e.Cancel Then MessageBox.Show("Zulässig sind nur Zahlen, mit maximal 2 Nachkommastellen", "Eingabefehler", MessageBoxButtons.OK)
   End Sub

   Public Shared Sub GridDelete_KeyDown(ByVal sender As Object, ByVal e As KeyEventArgs)
      If e.KeyData <> Keys.Delete Then Return
      Dim grd = DirectCast(sender, DataGridView)
      If grd.SelectedRows.Count > 0 Then Return
      e.Handled = True
      For Each c As DataGridViewCell In grd.SelectedCells
         With c.OwningColumn
            If .ReadOnly OrElse .DataPropertyName = "" Then Continue For
            Dim rw = DirectCast(c.OwningRow.DataBoundItem, DataRowView).Row
            rw(.DataPropertyName) = rw.Table.Columns(.DataPropertyName).DefaultValue
         End With
      Next
   End Sub

   Public Shared GridHideZeroNegativeRed_CellPainting As DataGridViewCellPaintingEventHandler _
      = Sub(s, e) GridNumberPainting(DirectCast(s, DataGridView), e, True)
   Public Shared GridNegativeRed_CellPainting As DataGridViewCellPaintingEventHandler _
      = Sub(s, e) GridNumberPainting(DirectCast(s, DataGridView), e, False)

   Private Shared _ObservedTypes As IList(Of Type) = New Type() {GetType(Integer), GetType(Double), GetType(Decimal)}

   Private Shared Sub GridNumberPainting(ByVal grd As DataGridView, ByVal e As DataGridViewCellPaintingEventArgs, ByVal hideZero As Boolean)
      If e.RowIndex < 0 OrElse e.ColumnIndex < 0 OrElse e.Value Is Nothing Then Return
      If Not _ObservedTypes.Contains(e.Value.GetType) Then Return
      Dim val = CType(e.Value, Decimal)
      If Not hideZero AndAlso val = 0 Then Return
      If val > 0 Then Return
      e.Handled = True
      Dim mask = DataGridViewPaintParts.All Xor DataGridViewPaintParts.ContentForeground
      e.Paint(e.CellBounds, e.PaintParts And mask)
      If val = 0 Then Return
      Dim col = grd.Columns(e.ColumnIndex)
      Dim rct = e.CellBounds
      With e.CellStyle.Padding
         rct.X += .Left : rct.Y += .Top : rct.Width -= .Horizontal : rct.Height -= .Vertical
      End With
      TextRenderer.DrawText(e.Graphics, e.FormattedValue.ToString, e.CellStyle.Font, rct, Color.Red, TextFormatFlags.Right Or TextFormatFlags.VerticalCenter)
   End Sub

   Public Shared NumberControl_TextChanged As EventHandler _
      = Sub(ctl, e)
           With DirectCast(ctl, Control)
              If .Text = "" Then Return
              .ForeColor = If(Single.Parse(.Text) < 0, Color.Red, SystemColors.ControlText)
           End With
        End Sub

   'Private Shared Sub Grid_EditingControlShowing(ByVal sender As Object, ByVal e As DataGridViewEditingControlShowingEventArgs)
   '   Dim grd = DirectCast(sender, DataGridView)
   '   With TryCast(e.Control, DataGridViewTextBoxEditingControl)
   '      If .Null Then
   '         _DroppingCombo = TryCast(grd.EditingControl, DataGridViewComboBoxEditingControl)
   '         If _DroppingCombo.Null Then Return
   '         AddHandler Application.Idle, AddressOf Application_Idle
   '      ElseIf grd.Columns(grd.CurrentCellAddress.X).DataPropertyName = "Vorgang" Then
   '         .AutoCompleteMode = AutoCompleteMode.SuggestAppend
   '         .AutoCompleteSource = AutoCompleteSource.CustomSource
   '         .AutoCompleteCustomSource = Program.Dts.AutoVorgang.Completions
   '      Else
   '         .AutoCompleteMode = AutoCompleteMode.None
   '      End If
   '   End With
   'End Sub

   ''' <summary> makes shure you can't enter a readonly Column at all. </summary>
   Public Shared Sub InitGrids(ByVal frm As Form)
      Dim getchhilds As Func(Of Control, IEnumerable) = Function(ctl) ctl.Controls
      Dim csDecimal = New DataGridViewCellStyle _
         With {.Alignment = DataGridViewContentAlignment.MiddleRight, .Format = "0.00#"}
      For Each grd In getchhilds.All(frm).OfType(Of DataGridView)()
         With grd
            .RowHeadersWidth = 25
            '.InitLockColumns()
            Try
               .ColumnHeadersDefaultCellStyle.Alignment = DataGridViewContentAlignment.MiddleCenter
            Catch ex As ArgumentException When ex.Message = "Spalte 'KasseName' gehört nicht zu Tabelle Kasse."
               'der komische Fehler tritt nur im Admin-Form auf, nur bei grdIndividuell. Habe ich jetzt umgangen, indem schon im Designer der 
               Dbg("Komischer Fehler: ", ex.Message)
            End Try
            If Not .ReadOnly AndAlso .ColumnCount > 0 Then
               AddHandler .CellValidating, AddressOf ControlBehavior.CurrencyGrid_CellValidating
               AddHandler .KeyDown, AddressOf ControlBehavior.GridDelete_KeyDown
            End If
            '.SelectionMode = DataGridViewSelectionMode.RowHeaderSelect
            .DefaultCellStyle.SelectionForeColor = SystemColors.ControlText
            AddHandler .DataError, AddressOf ControlBehavior.Grid_DataError
            AddHandler .GotFocus, AddressOf ControlBehavior.Grid_FocusChanged
            AddHandler .LostFocus, AddressOf ControlBehavior.Grid_FocusChanged
            ControlBehavior.Grid_FocusChanged(grd, EventArgs.Empty)
            Dim bs = DirectCast(.DataSource, BindingSource)
            If bs.Null Then Return
            Dim props = bs.GetItemProperties(Nothing)
            'Dim p2 = props.ToArray(Of PropertyDescriptor)()
            Dim handlesEditShowing = 0
            For Each col As DataGridViewColumn In .Columns
               If TypeOf col Is DataGridViewTextBoxColumn Then
                  If props(col.DataPropertyName).PropertyType.Equals(GetType(Decimal)) Then
                     If col.DefaultCellStyle.Format = "" Then col.DefaultCellStyle = csDecimal
                  ElseIf col.DataPropertyName = "Vorgang" Then
                     handlesEditShowing += 1
                  End If
               ElseIf TypeOf col Is DataGridViewComboBoxColumn Then
                  handlesEditShowing += 1
               End If
               If handlesEditShowing = 1 Then
                  'AddHandler grd.EditingControlShowing, AddressOf Grid_EditingControlShowing
               End If
            Next
         End With
      Next
   End Sub
End Class
