Imports System.Data.Common
Imports DbGenerator.DbGeneratorDts

Partial Class DbGeneratorDts

   Partial Class MappingDataTable

      Private WithEvents _Active As MappingDataTable = Me
      Public ReadOnly EventDisabler As New ActionFrame(Sub() _Active = Nothing, Sub() _Active = Me)

      Private Sub _Active_ColumnChanged(sender As Object, e As DataColumnChangeEventArgs) Handles _Active.ColumnChanged
         If e.Row.RowState = DataRowState.Detached OrElse e.Column IsNot PreferredColumn Then Return
         Dim rw = DirectCast(e.Row, MappingRow)
         If rw.IsFwNameNull Then Return
         Using EventDisabler.Enter
            Dim fwName = rw.FwName
            ' unchecking Preferred-Property of the other MappingRows
            Call (From mp In rw.ProviderRow.GetMappingRows _
               Where mp.Preferred AndAlso Not mp.IsFwNameNull AndAlso mp.FwName = fwName AndAlso mp IsNot rw) _
               .ForEach(Sub(r) r.Preferred = False)
         End Using
      End Sub

      Private Sub _Active_ColumnChanging(sender As Object, e As DataColumnChangeEventArgs) Handles _Active.ColumnChanging
         If e.Row.RowState = DataRowState.Detached Then Return
         Dim rw = DirectCast(e.Row, MappingRow)
         If rw.IsFwNameNull Then Return
         Select Case True
            Case e.Column Is PreferredColumn
               e.ProposedValue = True 'only can click to checked - unchecking the others does _Active_ColumnChanged()
            Case e.Column Is ParamDefaultValuesColumn
               If rw.ParamNames = "" Then e.ProposedValue = ""
         End Select
      End Sub

      Private preferenceKeys As String() = "image longbinary decimal numeric nvar var int".Split
      Private Function GetPreferable(candidates As MappingRow()) As MappingRow
         'a heuristic to get plausible Mapping-Preferences from the DbName of the Mappings
         If candidates.Length > 1 Then
            For Each key In preferenceKeys
               Dim candidate = candidates.FirstOrDefault(Function(rw) rw.DbName.IndexOf(key, StringComparison.OrdinalIgnoreCase) >= 0)
               If candidate.NotNull Then Return candidate
            Next
         End If
         Return candidates(0)
      End Function

      Friend Sub EnsurePreferences()
         Using EventDisabler.Enter
            'group MappingRows by Provider and FwName. Ensure, that each Group contains exact one "Preferred" MappingRow
            For Each grpProviderFwName In From rw In Me Where 0 = (rw.RowState And (DataRowState.Deleted Or DataRowState.Detached)) _
                  AndAlso Not rw.IsFwNameNull Group By rw.ProviderName, rw.FwName Into Group Select Group
               Dim grpPreferreds = From rw In grpProviderFwName Group By rw.Preferred Into Group Order By Preferred Select DirectCast(Group, MappingRow())
               Dim candidates = grpPreferreds.Last
               If candidates(0).Preferred Then
                  If candidates.Length = 1 Then Continue For 'no problem: exact one "Preferred" MappingRow
                  'problem: more than one "Preferreds" 
                  candidates.Except(GetPreferable(candidates)).ForEach(Sub(rw) rw.Preferred = False)
               Else 'problem: no-one "Preferred" 
                  GetPreferable(candidates).Preferred = True
               End If
            Next
         End Using
      End Sub

      Private Sub _Active_MappingRowChanged(sender As Object, e As MappingRowChangeEvent) Handles _Active.MappingRowChanged, _Active.MappingRowDeleted
         If (e.Action And (DataRowAction.Add Or DataRowAction.Delete)) = 0 Then Return
         EnsurePreferences()
      End Sub
   End Class 'MappingDataTable

End Class