﻿Imports System.Data.Common
Imports System.Runtime.CompilerServices
Imports DbGenerator.DbGeneratorDts

Partial Class DbGeneratorDts

   Partial Class ProjectRow

      Public Function ProposeSql() As String()
         'create sql to build up database fitting to the source-Dataset
         Dim lines = New List(Of String)
         Dim sl As New List(Of String)
         Dim addRuleActionClause As Action(Of String, Rule) = _
            Sub(ruleName As String, rule As Rule)
               If rule = Rule.None Then Return
               sl.Add("ON", ruleName, If(rule = Rule.Cascade, "CASCADE", If(rule = Rule.SetDefault, "SET DEFAULT", "SET NULL")))
            End Sub
         Dim uniqueName As New Uniquifier
         Dim columnHasNoExpression As Func(Of DataColumn, Boolean) = Function(col) col.Expression = ""
         Dim autoIncrementSyntax = ProviderRow.AutoIncrement.ToUpper
         Dim dts = DirectCast(Table.DataSet, DbGeneratorDts)
         Dim columnMappings = GetColumnMappingRows.ToDictionary(Function(r) r.TableName.And(".", r.Column))
         Using _SourceDts = New DataSet
            _SourceDts.ReadXmlSchema(Dataset)
            Dim tbls = _SourceDts.GetRankedTables
            For Each tbl As DataTable In tbls
               'per table: create table, columns, unique-constraints, foreignkey-constraints
               sl.Clear()
               sl.Add("CREATE TABLE", tbl.TableName.Quote, "(")
               For Each col In tbl.Columns.Cast(Of DataColumn).Where(columnHasNoExpression)
                  With col
                     Dim cmp = columnMappings(tbl.TableName.And(".", .ColumnName))
                     Dim mp = cmp.MappingRow
                     Dim dbName = mp.DbName
                     sl.Add(.ColumnName.Quote)
                     If .AutoIncrement Then
                        sl.Add(autoIncrementSyntax.Replace("?", dbName))
                     Else
                        sl.Add(If(dbName.LeftCut("("), dbName))
                        If mp.ParamNames <> "" Then
                           sl.Add(If(.MaxLength < 0, cmp.ParamValues, .MaxLength.ToString).InParenthesis)
                        End If
                     End If
                     If Not .AllowDBNull Then sl.Add("NOT NULL")
                  End With
                  sl.Add(",")
               Next
               For Each cnstr In tbl.Constraints.OfType(Of UniqueConstraint)()
                  With cnstr
                     If .IsPrimaryKey Then
                        If autoIncrementSyntax.Contains("PRIMARY KEY") Then Continue For
                        sl.Add("CONSTRAINT", ("PK_" & .Table.TableName).Quote, "PRIMARY KEY")
                     Else
                        sl.Add("CONSTRAINT", uniqueName("U_", .Table.TableName).Quote, "UNIQUE")
                     End If
                     sl.Add(.Columns.ColumngroupString)
                  End With
                  sl.Add(",")
               Next
               For Each cnstr In tbl.Constraints.OfType(Of ForeignKeyConstraint)()
                  With cnstr
                     sl.Add("CONSTRAINT", uniqueName("FK_", .RelatedTable.TableName, "_", .Table.TableName).Quote)
                     sl.Add("FOREIGN KEY", .Columns.ColumngroupString)
                     sl.Add("REFERENCES", .RelatedTable.TableName.Quote, .RelatedColumns.ColumngroupString)
                     addRuleActionClause("UPDATE", .UpdateRule)
                     addRuleActionClause("DELETE", .DeleteRule)
                  End With
                  sl.Add(",")
               Next
               sl.Poke(")")
               lines.Add(" ".Between(sl))
               lines.Add("")
            Next
            'aditional create index on each unique-constraint and foreignkey-parent-constraint
            Dim allUniques = From tbl In tbls
                                 From cnstr In tbl.Constraints.OfType(Of UniqueConstraint)()
                                 Where cnstr.Columns.All(columnHasNoExpression)
                                 Select tbl = tbl.TableName, columns = cnstr.Columns.ColumngroupString
            Dim allForeignKeyParents = From tbl In tbls
                                 From cnstr In tbl.Constraints.OfType(Of ForeignKeyConstraint)()
                                 Where cnstr.RelatedColumns.All(columnHasNoExpression)
                                 Select tbl = cnstr.RelatedTable.TableName, columns = cnstr.RelatedColumns.ColumngroupString
            For Each indx In allUniques.Concat(allForeignKeyParents).Distinct
               lines.Add("CREATE INDEX ".And(uniqueName("index" & indx.tbl).Quote, " ON ", indx.tbl.Quote & indx.columns))
            Next
            Return lines.ToArray
         End Using
      End Function

      Public Sub TryEnsureMappings()
         If ConnectionString = "" OrElse ProviderRow.IsNullRow OrElse ProviderRow.GetMappingRowsValid().Length > 0 Then Return
         ParseDbSchema()
      End Sub

      Public Sub ParseDbSchema()
         'read from Database-Schema QuoteIdentifiers and 5 Properties of each of the supported DataType
         Using con = CreateConnection()
            con.Open()
            Using tb = con.GetSchema("DatasourceInformation")
               Dim _Options As String() = {"[ ]", "` `"}
               Dim quotedIdentifierPattern = tb(0)("QuotedIdentifierPattern").ToString
               Dim c = quotedIdentifierPattern(quotedIdentifierPattern.IndexOf("|"c) + 2)
               Dim i = _Options.FindIndex(Function(s) s(2) = c)
               If i < 0 Then
                  Msg("could not read QuoteIdentifier reliable!")
                  ProviderRow.Quotes = String.Concat(c, " "c, c)
               Else
                  ProviderRow.Quotes = _Options(i)
               End If
            End Using
            Using tb = con.GetSchema("DataTypes")
               Dim dts = DirectCast(Table.DataSet, DbGeneratorDts)
               Using dts.Mapping.EventDisabler.Enter
                  ProviderRow.GetMappingRowsValid.ForEach(Sub(rwProv) rwProv.Delete())
                  For Each datarow As DataRow In tb.Rows
                     Dim rw As Func(Of String, String) = Function(s) datarow(s).ToString
                     Dim ParamNames = rw("CreateParameters"), CreateFormat = rw("CreateFormat"), _
                        dbName = rw("TypeName"), fwName = rw("DataType"), ColumnSize = rw("ColumnSize")
                     Dim paramSplits = ParamNames.Split({","c}, StringSplitOptions.RemoveEmptyEntries)
                     'maximal 5 ParamDefaultValues - der erste ist ColumnSize, alle weiteren "5"
                     Dim ParamDefaultValues = ", ".Between((ColumnSize & " 5 5 5 5").Split.Zip(paramSplits, Function(s1, s2) s1))
                     If CreateFormat <> "" Then
                        dbName = CreateFormat
                     ElseIf ParamNames <> "" Then   'optionale Params bei ohne CreateFormat: zunächstmal zusätzliche Mappingrow ohne Params einschieben
                        dts.Mapping.AddMappingRow(ProviderRow, dbName, fwName, False, "", "", ColumnSize, False)
                        dbName &= ",".Between(Enumerable.Range(0, 5).Zip(paramSplits, Function(i, s) "{".And(i, "}"))).InParenthesis
                     End If
                     dts.Mapping.AddMappingRow(ProviderRow, dbName, fwName, False, ParamNames, ParamDefaultValues, ColumnSize, False)
                  Next
                  dts.Mapping.EnsurePreferences()
               End Using
            End Using
         End Using
      End Sub

      Public Sub CreateColumnMappings()
         'a ColumnMappingRow associates a DataTable-+Column-Name with the for its DataType preferred MappingRow - otherwise the Providers NullMappingRow
         GetColumnMappingRows().ForEach(Sub(rw) rw.Delete())
         Dim mappings = ProviderRow.GetMappingRows
         Dim nullMapping = mappings.First(Function(mp) mp.IsNullRow)
         Dim dicMaps = mappings.Where(Function(rw) rw.Preferred).ToDictionary(Function(rw) rw.FwName)
         Dim dts = DirectCast(Table.DataSet, DbGeneratorDts)
         Using _SourceDts = New DataSet
            _SourceDts.ReadXmlSchema(Dataset)
            For Each tbl As DataTable In _SourceDts.GetRankedTables
               For Each clm As DataColumn In tbl.Columns
                  Dim mp = If(dicMaps.GetOrDefault(clm.DataType.ToString), nullMapping)
                  dts.ColumnMapping.AddColumnMappingRow(tbl.TableName, clm.ColumnName, clm.DataType.ToString, Me, mp, mp.ParamDefaultValues)
               Next
            Next
         End Using
      End Sub

      Public Sub ClearDB()
         Using con = CreateConnection(), cmd = con.CreateCommand
            con.Open()
            ClearDB(con, cmd)
         End Using
      End Sub

      Private Sub ClearDB(con As DbConnection, cmd As DbCommand)
         'untergeordnete Tabellen müssen zuerst gedropt werden, und es gibt auch geschützte Tabellen. Ich aber versuche immer wieder wahllos alle Tabellen zu droppen, das löscht pro Durchlauf zumindest die unterste Ebene. Ein Durchlauf ganz ohne erfolgreichen Drop beendet die Methode.
         Using tb = con.GetSchema("Tables")
            Dim dropCommands = (From rw In tb.AsEnumerable Select "DROP TABLE ".And(rw("TABLE_NAME").ToString.Quote)).ToArray
            Dim success = True
            While success
               success = False
               For i = dropCommands.Length - 1 To 0 Step -1
                  If dropCommands(i) = "" Then Continue For
                  Try
                     cmd.CommandText = dropCommands(i)
                     cmd.ExecuteNonQuery()
                     success = True
                     dropCommands(i) = ""
                  Catch ex As Exception
                  End Try
               Next
            End While
         End Using
      End Sub

      Public Sub ExecuteSql(sqlBuildUp As IEnumerable(Of String))
         Using con = CreateConnection(), cmd = con.CreateCommand
            con.Open()
            ClearDB(con, cmd)
            ExecuteCommands(cmd, sqlBuildUp, True) 'execute sqlBuildUp
         End Using
      End Sub

      Private Sub ExecuteCommands(cmd As DbCommand, sql As IEnumerable(Of String), display As Boolean)
         'on display: display each line and rethrow errors - otherwise keep silent even in case of exceptions
         'swallowing exceptions is useful while clearing database, since i don't consider drop-protection of database-objects
         For Each line In sql.Where(Function(ln) ln.Length > 3)
            Try
               cmd.CommandText = line
               cmd.ExecuteNonQuery()
               line = "Success".And(Lf, line)
            Catch ex As Exception
               'on rethrow include error-line and InnerException, if available
               If display Then Throw Me.Exception(If(ex.InnerException.Null, ex, ex.InnerException), line)
            End Try
            If display Then Msg(line)
         Next
      End Sub
   End Class 'ProjectRow
End Class

Module QuoteExtensions
   Dim _InnerQuotes As String
   Dim _OuterQuotes As String()
   Dim _Parenthesis As String() = {"( ", " )"}

   Sub InitQuotes(quotes As String)
      _OuterQuotes = quotes.Split
      _InnerQuotes = _OuterQuotes(1).And(",", _OuterQuotes(0))
   End Sub
   <Extension()> _
   Function ColumngroupString(cols As DataColumn()) As String
      Return _InnerQuotes.Between(cols).Between(_OuterQuotes).InParenthesis
   End Function
   <Extension()> _
   Public Function InParenthesis(ByVal Subj As String) As String
      Return Subj.Between(_Parenthesis)
   End Function
   <Extension()> _
   Public Function Quote(ByVal Subj As String) As String
      Return Subj.Between(_OuterQuotes)
   End Function
   <Extension()> _
   Function GetFactory(rw As ProjectRow) As DbProviderFactory
      If rw.Null Then
         Throw New ArgumentNullException("rw", "scheinbar ist noch kein Projekt geladen")
      End If
      Select Case rw.ProviderRow.InvariantName
         Case "System.Data.SQLite"
            Return SQLite.SQLiteFactory.Instance ' fails (on my system)
         Case "MySql.Data.MySqlClient"
            Return MySql.Data.MySqlClient.MySqlClientFactory.Instance
         Case Else
            Return DbProviderFactories.GetFactory(rw.ProviderRow.InvariantName)
      End Select
   End Function
   <Extension()> _
   Function CreateConnection(rw As ProjectRow) As DbConnection
      CreateConnection = rw.GetFactory.CreateConnection
      CreateConnection.ConnectionString = rw.ConnectionString
   End Function
End Module