﻿Imports System.Data.Common
Imports DbGenerator.DbGeneratorDts
Imports System.IO
Imports System.ComponentModel

Public Class frmMain

   Public Sub New()
      InitializeComponent()
      ControlNotifying()
      Using DbGeneratorDts.Mapping.EventDisabler.Enter
         DbGeneratorDts.Register(Me, True).Fill()
         EnsureMinimumData()
      End Using
      ControlBehavior.InitGrids(Me)
      ColumnMappingDataGridView.Attach.DrawGroupLines({DbGeneratorDts.ColumnMapping.TableNameColumn})
      AdvancedBindings()
   End Sub

   Private Sub ControlNotifying()
      'some Checkboxes change visibility of some other controls
      Dim Inverter As Converter(Of Boolean, Boolean) = Function(b) Not b
      ckProjects.Property("Checked").Notify(spltcntProjects, "Panel1Collapsed", Inverter).Notify(btParseProject, "Visible")
      ckMappings.Property("Checked").Notify(spltcntMappingsSql, "Panel1Collapsed", Inverter)
      With ckSql.Property("Checked")
         For Each bt In {btClearDB, btProposeSql, btExecuteSql}
            bt.Visible = ckSql.Checked : .Notify(bt, "Visible")
         Next
         .OnInit().Notify(spltcntMappingsSql, "Panel2Collapsed", Inverter)
      End With
   End Sub

   Private Sub EnsureMinimumData()
      'enshure a Nullrow to Provider, which means, that no Provider is selected
      With DbGeneratorDts.Provider.NewProviderRow
         .Name = "Kein"
         .InvariantName = "Kein"
         .SetQuotesNull()
         .EnsureExistUnique(DbGeneratorDts.Provider.QuotesColumn)
         DbGeneratorDts.Project.ProviderNameColumn.DefaultValue = .Name
      End With
      For Each rw As DataRow In DbProviderFactories.GetFactoryClasses.Rows
         'enshure ProviderRows according to each installed Provider
         Dim rwProv = DbGeneratorDts.Provider.NewProviderRow
         rwProv.InvariantName = rw("InvariantName").ToString
         rwProv.Name = rwProv.InvariantName.LastRightCut("Data.")
         rwProv.EnsureExistUnique(DbGeneratorDts.Provider.InvariantNameColumn)
      Next
      For Each rwProv In DbGeneratorDts.Provider
         'enshure a subordered MappingRow as Nullrow, which means: in this Provider no Mapping is selected
         With DbGeneratorDts.Mapping
            Dim rwNull = .NewMappingRow()
            rwNull.ProviderRow = rwProv
            rwNull.EnsureExistUnique(.FwNameColumn, .ProviderNameColumn)
            If rwNull.CanDelete Then rwNull.CanDelete = False
         End With
      Next
   End Sub

   Private Sub AdvancedBindings()
      Dim bnd = lbDataset.DataBindings.Add("Text", UclProjects.srcProject, "Dataset", True, DataSourceUpdateMode.Never)
      AddHandler bnd.Format, Sub(s, e) e.Value = If(e.Value.ToString = "", "<kein>", Path.GetFileName(e.Value.ToString))
      srcColumnMapping.ChangeBinding(UclProjects.srcProject, "Project_ColumnMapping")
      UclProjects.srcProject.BindRelatedSource(UclProviderMappings.srcProvider)
      srcColumnMapping.BindRelatedSource(UclProviderMappings.srcMappings)
   End Sub

   Private Sub AnyButton_Click(sender As Object, e As EventArgs) Handles btProviderSettings.Click, btSaveSettings.Click, btParseProject.Click, btDatabaseViewer.Click, btProposeSql.Click, btExecuteSql.Click, btClearDB.Click, btAddMapping.Click, btReloadSettings.Click, TestToolStripMenuItem.Click
      Dim rwProj = UclProjects.srcProject.At(Of ProjectRow)()
      Select Case True
         Case sender Is TestToolStripMenuItem
            rwProj.ParseDbSchema()

         Case sender Is btSaveSettings : DbGeneratorDts.Save(Me) : Return
         Case sender Is btReloadSettings
            Using DbGeneratorDts.Mapping.EventDisabler.Enter
               DbGeneratorDts.Fill()
            End Using

         Case sender Is btClearDB : rwProj.ClearDB()
         Case sender Is btParseProject : rwProj.CreateColumnMappings()
         Case sender Is btProposeSql : txtSql.Lines = rwProj.ProposeSql
         Case sender Is btExecuteSql
            Try
               rwProj.ExecuteSql(txtSql.Lines)
            Catch ex As Exception(Of ProjectRow)
               Dim txt = Lf.Between(Lf, ex.Message, ex.InnerException.Message)
               txtSql.AppendText(txt)
               MessageBox.Show(txt, ex.InnerException.GetType.Name, MessageBoxButtons.OK, MessageBoxIcon.Error)
            End Try
         Case sender Is btProviderSettings
            frmProvider.Show(Me)
            Return
         Case sender Is btDatabaseViewer
            Try
               Call New frmDataBaseViewer(rwProj.GetFactory, rwProj.ConnectionString).Show()
            Catch ex As DbException
               Msg(rwProj.ConnectionString, Lf, ex.Message)
            End Try
            Return
         Case sender Is btAddMapping
            Dim defaultValues = TupleList.From(Of DataColumn, Object)(DbGeneratorDts.Mapping.FwNameColumn, srcColumnMapping.At(Of ColumnMappingRow).FwName)
            UclProviderMappings.srcMappings.EditNew(Of dlgMapping)(defaultValues)
            Return
      End Select
      Media.SystemSounds.Asterisk.Play()
   End Sub

   Private Sub ckMappings_CheckedChanged(sender As Object, e As EventArgs) Handles ckMappings.CheckedChanged, ckSql.CheckedChanged
      'verhindern dass beide Checkboxen unchecked werden, denn die daran hängenden SplitterPanel können nicht gleichzeitig collapsed werden.
      Dim cks = {ckMappings, ckSql}
      Dim iSender = cks.IndexOf(DirectCast(sender, CheckMenu))
      If Not cks(iSender).Checked Then cks(iSender Xor 1).Checked = True
   End Sub

   Private Sub ColumnMappingDataGridView_CellPainting(sender As Object, e As DataGridViewCellPaintingEventArgs) Handles ColumnMappingDataGridView.CellPainting
      'alarm-color, if ColumnMappingRow refers to a NullMappingRow.
      If e.ColumnIndex <> clmCmbMappingID.DisplayIndex OrElse Not e.RowIndex.IsBetween(0, srcColumnMapping.Count) Then Return
      If (e.State And DataGridViewElementStates.Selected) <> 0 Then Return
      If Not srcColumnMapping.At(Of ColumnMappingRow)(e.RowIndex).MappingRow.IsNullRow Then Return
      e.Graphics.FillRectangle(Brushes.OrangeRed, e.CellBounds)
      e.Paint(e.CellBounds, DataGridViewPaintParts.All Xor DataGridViewPaintParts.Background)
      e.Handled = True
   End Sub

End Class