﻿Imports System.Runtime.CompilerServices
Imports System.Reflection
Imports System.ComponentModel
Imports System.Data
Imports System.IO

Namespace System.Windows.Forms
   Public Module XDatasetAdapter
      <Extension()> _
      Public Function SuspendConstraints(dts As DataSet) As IDisposable
         Return dts.Attach.SuspendConstraints.Enter
      End Function
      <Extension()> _
      Public Sub RegisterSaveCallback(dts As DataSet, dlg As Action, Optional value As Boolean = True)
         dts.Attach.RegisterSaveCallback(dlg, value)
      End Sub
      <Extension()> _
      Public Function Attach(dts As DataSet) As DataTableAttach.DatasetAttach
         Return dts.Tables(0).Attach.Dataset_Attach
      End Function
      <Extension()> _
      Public Function Adapter(dts As DataSet) As DatasetAdapterBase
         Adapter = Nothing
         If DatasetAdapterBase.Attachs.TryGetValue(dts, Adapter) Then Exit Function
         Adapter = New DatasetXmlAdapter(dts)
         DatasetAdapterBase.Attachs.Add(dts, Adapter)
      End Function
      <Extension()> _
      Public Function DataFile(dts As DataSet, path As String) As DataSet
         DirectCast(dts.Adapter, DatasetXmlAdapter).DataFile = New FileInfo(path)
         Return dts
      End Function
      <Extension()> _
      Public Function DataFile(dts As DataSet) As FileInfo
         Dim att = DirectCast(dts.Adapter, DatasetXmlAdapter)
         If att.Null Then Return Nothing
         Return att.DataFile
      End Function

      ''' <summary>Clear dataset and refill. Optional only the given tables (not supported in dataset-only-scenarios)</summary>
      <Extension()> _
      Public Sub Fill(dts As DataSet, ParamArray tables() As DataTable)
         dts.Adapter.Fill(tables)
      End Sub
      <Extension()> _
      Public Sub Save(dts As DataSet, frm As Form, Optional checkForChanges As Boolean = True)
         dts.Adapter.Save(frm, checkForChanges)
      End Sub

      <Extension()> _
      Public Function HandleFormClosing(dts As DataSet) As FormClosingEventHandler
         Return dts.Adapter.HandleFormClosing
      End Function

#Region "Winfom-Close-Bug"         ' Aufwändig gestaltete DGVs schmeißen manchmal Fehler beim Disposen des Forms.

      Private Sub Form_Closed(s As Object, e As EventArgs)
         'manchmal schmeissen DGVs Fehler beim Schließen des Forms. Ist schlecht reproduzierbar.
         'Reproduziere: Starte RaslDb, gehe auf Artikel-Tab und Form schließen
         With DirectCast(s, Form)
            If .DialogResult <> DialogResult.None Then Return
            .Validate() 'nochn Bug: DGV kriegt beim Disposen einen ZeilenIndex-Fehler, wenns sich die ZufügeZeile im Edit-Modus befindet
            .SuspendLayout()
#If True Then            'geringster Eingriff
            For Each ctl In New GetChilds(Of Control)(Function(c) c.Controls).AllAsList(.Self)
               With TryCast(ctl, DataGridView)
                  If .NotNull Then .Parent.SuspendLayout() : .Dispose()
               End With
            Next
#ElseIf True Then            'einfachster Eingriff
            With .Controls
               'Controls könnten sich theor. auch gegenseitig disposen, über Events. dieses hier disposed ganz sicher immer das letzte Element was noch da ist
               While .Count > 0 : .Item(.Count - 1).Dispose() : End While
            End With
#ElseIf True Then            'kein Eingriff
         Return
#ElseIf True Then
#End If
         End With
      End Sub

      'Beim Schließen des MainForms wird Form_Closed der anneren Forms garnet mehr aufgerufen. Daher hier alle alle Controls disposen
      Private Sub MainForm_Closed(s As Object, e As EventArgs)
         For Each frm As Form In Application.OpenForms
            Form_Closed(frm, e)
         Next
      End Sub

#End Region 'Winfom-Close-Bug

      Private _KnownDatasets As New Dictionary(Of Type, DataSet)

      ''' <summary>
      ''' registers all datasets of this type and all bindingsources of the form, to improve performance and to avoid data-redundance
      ''' </summary>
      <Extension()> _
      Public Function Register(Of T As DataSet)(dts As T, frm As Form, handleFormClosing As Boolean) As T
         Dim tpDts = dts.GetType
         Dim tmp As DataSet = Nothing
         If _KnownDatasets.TryGetValue(tpDts, tmp) Then
            dts = DirectCast(tmp, T)
         Else
            _KnownDatasets.Add(tpDts, dts)
         End If
         Register = dts
         AddHandler frm.FormClosed, AddressOf Form_Closed
         If handleFormClosing Then AddHandler frm.FormClosing, dts.Adapter.HandleFormClosing
         Dim allCtls = New GetChilds(Of Control)(Function(ctl) ctl.Controls).AllAsList(frm)
         Dim dgvErrorSuspend = New GridErrorSuspend(dts, allCtls)
         Dim oldDatasets As New HashSet(Of DataSet)
         Dim shouldReplace As Func(Of Object, Boolean) = Function(src) _
            src IsNot Nothing AndAlso _
            (src Is tpDts OrElse src.GetType Is tpDts AndAlso src IsNot dts)
         Dim bindFlags As BindingFlags = BindingFlags.Instance Or BindingFlags.Public _
            Or BindingFlags.NonPublic Or BindingFlags.GetField
         For Each ctl In allCtls.OfType(Of ContainerControl)()
            For Each fld In ctl.GetType.GetFields(bindFlags)
               If fld.FieldType Is tpDts Then
                  oldDatasets.Add(DirectCast(fld.GetValue(ctl), DataSet))
                  fld.SetValue(ctl, dts)
                  Continue For
               End If
               If Not GetType(IComponent).IsAssignableFrom(fld.FieldType) Then Continue For
               Dim prp = fld.FieldType.GetProperty("DataSource")
               If prp.Null Then Continue For
               Dim itm = fld.GetValue(ctl)
               If itm Is Nothing Then Continue For
               TryRegisterAsBindingSource(itm, dts, tpDts)
               If shouldReplace(prp.GetValue(itm, Nothing)) Then prp.SetValue(itm, dts, Nothing)
            Next
         Next
         dgvErrorSuspend.Resume()
         oldDatasets.Remove(dts)
         For Each dts In oldDatasets : dts.Dispose() : Next
      End Function

      Private Sub TryRegisterAsBindingSource(itm As Object, dts As DataSet, tpDts As Type)
         Dim bs = TryCast(itm, BindingSource)
         If bs.Null Then Return
         Dim dataMember = bs.DataMember
         Dim dataSource = bs.DataSource
         Do
            With TryCast(dataSource, BindingSource)
               If .Null Then Exit Do
               If dataMember = "" Then dataMember = .DataMember
               dataSource = .DataSource
            End With
         Loop
         If dataSource IsNot tpDts AndAlso dataSource.GetType IsNot tpDts Then Return
         With dts.Tables(dataMember)
            If .NotNull Then .Attach.AddSource(bs) : Return
         End With
         With dts.Relations(dataMember)
            If .NotNull Then .ChildTable.Attach.AddSource(bs)
         End With
      End Sub

      Private Class GridErrorSuspend
         'beim Umstöpseln der DataSources können in DGVs DataErrors auftreten. Diese werden suspendiert, wenn temporär ein  DataGridViewDataErrorEventHandler das dgv.DataError-Event behandelt 
         Private _Grids As New List(Of DataGridView)
         Private _DummiHandler As DataGridViewDataErrorEventHandler = Function() 0
         Public Sub New(dts As DataSet, allCtls As List(Of Control))
            If dts.Tables.Cast(Of DataTable).All(Function(tb) tb.Rows.Count = 0) Then Return
            For Each ctl In allCtls
               Dim dgv = TryCast(ctl, DataGridView)
               If dgv.Null Then Continue For
               AddHandler dgv.DataError, _DummiHandler
               _Grids.Add(dgv)
            Next
         End Sub
         Public Sub [Resume]()
            For Each dgv In _Grids
               Dim bs = TryCast(dgv.DataSource, BindingSource)
               If bs.NotNull Then bs.ResetBindings(False)
            Next
         End Sub
      End Class
   End Module
End Namespace

