﻿Imports System
Imports System.Drawing
Imports System.ComponentModel
Imports System.Windows.Forms.Layout
Namespace System.Windows.Forms

   ''' <summary>
   ''' override Control.LayoutEngine and return NoLayout.Instance to deactivate Bases LayoutEngine
   ''' </summary>
   Public Class NoLayout : Inherits LayoutEngine
      Public Shared ReadOnly Instance As New NoLayout
   End Class

   Public Module modBounds

      Private ReadOnly Boundses As New ComponentProperty(Of Control, BoundsEx) _
         (Function(ctl) New DetectingBoundsEx(ctl))

      <System.Runtime.CompilerServices.Extension()> _
      Public Function BoundsX(ByVal ctl As Control) As BoundsEx
         Return Boundses(ctl)
      End Function

      ''' <summary>
      ''' holds some Control-Properties (as Vector-Fields), which are important for userdefined Layout.
      ''' You can use these Fields to compute new Bounds without side-effect to the control. The Array
      ''' CanSize(1) As Boolean informes, in what direction the control is sizeable
      ''' </summary>
      Public Class DetectingBoundsEx : Inherits BoundsEx

         Public Sub New(ByVal ctl As Control)
            MyBase.New(ctl)
            AutoDetectCanSize()
         End Sub

         ''' <summary>
         ''' Caution! Try avoid to call this expensive method. Only call while 
         ''' handling Layout. Wrong use can cause flickering.
         ''' </summary>
         Public Overrides Sub AutoDetectCanSize()
            'check out, wether Ctl is sizeable, and in what Direction
            With Ctl
               Dim canS = Not .AutoSize
               CanSize(0) = canS : CanSize(1) = canS
               'only respect Ctl.AutoSize, if it is not attributed as EditorBrowsable.Never
               Dim attr = .GetType.GetProperty("AutoSize") _
                  .GetCustomAttributes(GetType(EditorBrowsableAttribute), True) _
                  .Cast(Of EditorBrowsableAttribute).FirstOrDefault
               If attr.Null OrElse attr.State <> EditorBrowsableState.Never Then Return
               ' worst case: try it out
               .Parent.SuspendLayout()
               .SuspendLayout()
               ReInit()
               Dim dlt = New Vector(50, 50)
               Dim temp1 = Size + dlt
               For i = 0 To 1
                  ' ensure .MaximumSize is big enough
                  If temp1(i) > Max(i) Then
                     .MaximumSize = Max + dlt
                     Exit For
                  End If
               Next
               .Size = temp1                                                              'change Ctl.Size
               Dim temp2 As Vector = .Size
               For i = 0 To 1
                  CanSize(i) = temp2(i) = temp1(i)                    'check if Size-changing was rejected
               Next
               ' re-set Ctl to previous Values
               .Size = Size
               .MaximumSize.Assign(Max.ToSize)
               .ResumeLayout(False)
               .Parent.ResumeLayout(False)
            End With
         End Sub
      End Class


      ''' <summary>
      ''' holds some Control-Properties (as Vector-Fields), which are important for userdefined Layout.
      ''' You can use these Fields to compute new Bounds without side-effect to the control. The Array
      ''' CanSize(1) As Boolean informes, in what direction the control is sizeable
      ''' </summary>
      <DebuggerDisplay("{Ctl.Name} {Location.X} {Location.Y} {Size.X} {Size.Y}")> _
      Public Class BoundsEx

         Public Shared ControlMin As New Vector(20, 13)
         Public Shared ControlDefault As New Vector(50, 20)

         Public Location As Vector
         Public Size As Vector
         Public Margin1 As Vector 'left/top - Margin
         Public Margin2 As Vector 'right/bottom - Margin
         Public Min As Vector '.MinimumSize
         Public Max As Vector '.MaximumSize
         ''' <summary>indicates, in what direction the control is sizeable</summary>
         Public ReadOnly CanSize As Boolean() = New Boolean() {True, True}

         Public ReadOnly Ctl As Control

         ''' <summary>creates a BoundsEx with inactive CanSize-AutoDetection, 
         ''' and makes it accessible to the Control.BoundsX-Extension-Function. 
         ''' Such Controls must set their .BoundsX.CanSize-Property by themselfs
         ''' </summary>
         Public Shared Function CreateEntry(ByVal ctl As Control) As BoundsEx
            CreateEntry = New BoundsEx(ctl)
            Boundses(ctl) = CreateEntry
         End Function

         Protected Sub New(ByVal ctl As Control)
            Me.Ctl = ctl
            If ctl.Parent.NotNull Then ReInit()
         End Sub

         Public ReadOnly Property MarginSize() As Vector
            Get
               Return Margin1 + Margin2
            End Get
         End Property

         Public Sub ReInit()
            With Ctl
               ' making up MiniMax, if needed: set .MinimumSize greater than ControlMin, replace
               '°.MaximumSize, if .IsEmpty, with SizeBig 
               If .Parent.Null Then
                  .MinimumSize = ControlMin.Union(.MinimumSize)
                  If .MaximumSize.IsEmpty Then .MaximumSize = SizeBig
               Else
                  'try avoid call ResumeLayout
                  Dim changeFlag As Integer = 0
                  Dim minimum = ControlMin.Union(.MinimumSize)
                  If minimum <> .MinimumSize Then changeFlag = 1
                  If .MaximumSize.IsEmpty Then changeFlag = changeFlag Or 2
                  If changeFlag > 0 Then
                     .Parent.SuspendLayout()
                     If (changeFlag And 1) > 0 Then .MinimumSize = minimum
                     If (changeFlag And 2) > 0 Then .MaximumSize = SizeBig
                     .Parent.ResumeLayout(False)
                  End If
               End If

               'Read Values from Control
               Location = .Location
               Size = .Size
               Min = .MinimumSize
               Max = .MaximumSize
               With .Margin
                  Margin1 = .Vector1
                  Margin2 = .Vector2
               End With
            End With
         End Sub

         ''' <summary> just a placeholder </summary>
         Public Overridable Sub AutoDetectCanSize()
         End Sub

         ''' <summary>TopLeft of controls rectangle, including Margins</summary>
         Public Property TopLeft() As Vector
            Get
               Return Location - Margin1
            End Get
            Set(ByVal value As Vector)
               Location = value + Margin1
            End Set
         End Property

         ''' <summary>controls Size, including Margins</summary>
         Public ReadOnly Property Fullsize(ByVal i As Integer) As Integer
            Get
               Return Size(i) + MarginSize(i)
            End Get
         End Property

         ''' <summary>
         ''' sets Size to a value the control will accept.
         ''' returns the difference between previous and new Size
         ''' </summary>
         Public Function ApplyLimitation() As Vector
            ApplyLimitation = Size
            Dim vCtl As Vector = Ctl.Size
            For i = 0 To 1
               Size(i) = If(CanSize(i), Size(i).ClipIn(Min(i), Max(i)), vCtl(i))
            Next
            ApplyLimitation -= Size
         End Function

         Public Sub ApplyBounds(ByVal checkLimitation As Boolean)
            If checkLimitation Then Me.ApplyLimitation()
            Ctl.Bounds = New Rectangle(Location, Size)
         End Sub

         ''' <summary>
         ''' returns the difference between proposedValue and the actually accepted value
         ''' </summary>
         Public Function TrySetFullSize(ByVal proposedValue As Integer, ByVal i As Integer) As Integer
            Return TrySetSize(proposedValue - MarginSize(i), i)
         End Function

         ''' <summary>
         ''' returns the difference between proposedValue and the actually accepted value
         ''' </summary>
         Public Function TrySetSize(ByVal proposedValue As Integer, ByVal i As Integer) As Integer
            If CanSize(i) Then Size(i) = proposedValue.ClipIn(Min(i), Max(i))
            Return proposedValue - Size(i)
         End Function

         Public Overrides Function ToString() As String
            Return String.Concat(Ctl.Name, Location, Size)
         End Function
      End Class

   End Module

End Namespace
